package dev.shadowsoffire.apotheosis.adventure.socket.gem;

import com.mojang.datafixers.kinds.Applicative;
import java.util.Collections;
import dev.shadowsoffire.placebo.codec.PlaceboCodecs;
import com.mojang.datafixers.kinds.App;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import java.util.Arrays;
import dev.shadowsoffire.apotheosis.adventure.affix.Affix;
import dev.shadowsoffire.placebo.reload.DynamicHolder;
import dev.shadowsoffire.apotheosis.adventure.socket.SocketHelper;
import net.minecraft.ChatFormatting;
import net.minecraft.network.chat.CommonComponents;
import net.minecraft.network.chat.Style;
import net.minecraft.network.chat.Component;
import java.util.function.Consumer;
import net.minecraft.world.item.ItemStack;
import java.util.Iterator;
import java.util.stream.Stream;
import java.util.Comparator;
import java.util.function.Predicate;
import java.util.Objects;
import dev.shadowsoffire.apotheosis.adventure.loot.RarityRegistry;
import java.util.function.Function;
import java.util.stream.Collectors;
import org.apache.commons.lang3.tuple.Pair;
import com.google.common.base.Preconditions;
import java.util.Optional;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import java.util.Map;
import javax.annotation.Nullable;
import dev.shadowsoffire.apotheosis.adventure.socket.gem.bonus.GemBonus;
import java.util.List;
import net.minecraft.resources.ResourceLocation;
import java.util.Set;
import com.mojang.serialization.Codec;
import dev.shadowsoffire.apotheosis.adventure.compat.GameStagesCompat;
import dev.shadowsoffire.apotheosis.adventure.loot.RarityClamp;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;
import dev.shadowsoffire.placebo.codec.CodecProvider;

public class Gem implements CodecProvider<Gem>, WeightedDynamicRegistry.ILuckyWeighted, WeightedDynamicRegistry.IDimensional, RarityClamp, GameStagesCompat.IStaged
{
    public static final Codec<Gem> CODEC;
    protected final int weight;
    protected final float quality;
    protected final Set<ResourceLocation> dimensions;
    protected final List<GemBonus> bonuses;
    protected final boolean unique;
    @Nullable
    protected final Set<String> stages;
    protected final transient Map<LootCategory, GemBonus> bonusMap;
    protected final transient int uuidsNeeded;
    protected final transient LootRarity minRarity;
    protected final transient LootRarity maxRarity;
    
    public Gem(final int weight, final float quality, final Set<ResourceLocation> dimensions, final Optional<LootRarity> minRarity, final Optional<LootRarity> maxRarity, final List<GemBonus> bonuses, final boolean unique, final Optional<Set<String>> stages) {
        this.weight = weight;
        this.quality = quality;
        this.dimensions = dimensions;
        this.bonuses = bonuses;
        this.unique = unique;
        this.stages = stages.orElse(null);
        Preconditions.checkArgument(!bonuses.isEmpty(), (Object)"No bonuses were provided.");
        this.bonusMap = bonuses.stream().mapMulti((gemData, mapper) -> {
            gemData.getGemClass().types().iterator();
            final Iterator iterator;
            while (iterator.hasNext()) {
                final LootCategory c = iterator.next();
                mapper.accept(Pair.of((Object)c, (Object)gemData));
            }
            return;
        }).collect(Collectors.toMap((Function<? super Object, ? extends LootCategory>)Pair::getLeft, (Function<? super Object, ? extends GemBonus>)Pair::getRight));
        this.uuidsNeeded = this.bonuses.stream().mapToInt(GemBonus::getNumberOfUUIDs).max().orElse(0);
        if (minRarity.isPresent()) {
            this.minRarity = minRarity.get();
        }
        else {
            final Stream stream = RarityRegistry.INSTANCE.getValues().stream();
            final GemBonus obj = bonuses.get(0);
            Objects.requireNonNull(obj);
            this.minRarity = (LootRarity)stream.filter(obj::supports).min(LootRarity::compareTo).get();
        }
        if (maxRarity.isPresent()) {
            this.maxRarity = maxRarity.get();
        }
        else {
            final Stream stream2 = RarityRegistry.INSTANCE.getValues().stream();
            final GemBonus obj2 = bonuses.get(0);
            Objects.requireNonNull(obj2);
            this.maxRarity = (LootRarity)stream2.filter(obj2::supports).max(LootRarity::compareTo).get();
        }
        Preconditions.checkArgument(this.minRarity.ordinal() <= this.maxRarity.ordinal(), (Object)"The min rarity must be <= the max rarity.");
    }
    
    public int getNumberOfUUIDs() {
        return this.uuidsNeeded;
    }
    
    public void addInformation(final ItemStack gem, final LootRarity rarity, final Consumer<Component> list) {
        if (this.isUnique()) {
            list.accept((Component)Component.m_237115_("text.apotheosis.unique").m_130948_(Style.f_131099_.m_178520_(13056274)));
            list.accept(CommonComponents.f_237098_);
        }
        final Style style = Style.f_131099_.m_178520_(720650);
        list.accept((Component)Component.m_237115_("text.apotheosis.socketable_into").m_130948_(style));
        addTypeInfo(list, this.bonusMap.keySet().toArray());
        list.accept(CommonComponents.f_237098_);
        list.accept((Component)Component.m_237115_("text.apotheosis.when_socketed_in").m_130940_(ChatFormatting.GOLD));
        for (GemBonus bonus : this.bonuses) {
            if (!bonus.supports(rarity)) {
                continue;
            }
            final Component modifComp = bonus.getSocketBonusTooltip(gem, rarity);
            final Component sum = (Component)Component.m_237110_("text.apotheosis.dot_prefix", new Object[] { Component.m_237110_("%s: %s", new Object[] { Component.m_237115_("gem_class." + bonus.getGemClass().key()), modifComp }) }).m_130940_(ChatFormatting.GOLD);
            list.accept(sum);
        }
    }
    
    public boolean canApplyTo(final ItemStack socketed, final ItemStack gem, final LootRarity rarity) {
        if (this.isUnique()) {
            final List<Gem> gems = SocketHelper.getGems(socketed).streamValidGems().map(GemInstance::gem).map((Function<? super Object, ? extends Gem>)DynamicHolder::get).toList();
            if (gems.contains(this)) {
                return false;
            }
        }
        return this.isValidIn(socketed, gem, rarity);
    }
    
    public boolean isValidIn(final ItemStack socketed, final ItemStack gem, final LootRarity rarity) {
        final LootCategory cat = LootCategory.forItem(socketed);
        return !cat.isNone() && this.bonusMap.containsKey(cat) && this.bonusMap.get(cat).supports(rarity);
    }
    
    public Optional<GemBonus> getBonus(final LootCategory cat, final LootRarity rarity) {
        return Optional.ofNullable(this.bonusMap.get(cat)).filter(b -> b.supports(rarity));
    }
    
    @Override
    public String toString() {
        return String.format("Gem: %s", this.getId());
    }
    
    public static String fmt(final float f) {
        return Affix.fmt(f);
    }
    
    public float getQuality() {
        return this.quality;
    }
    
    public int getWeight() {
        return this.weight;
    }
    
    public Set<ResourceLocation> getDimensions() {
        return this.dimensions;
    }
    
    public LootRarity getMaxRarity() {
        return this.maxRarity;
    }
    
    public LootRarity getMinRarity() {
        return this.minRarity;
    }
    
    public List<GemBonus> getBonuses() {
        return this.bonuses;
    }
    
    public boolean isUnique() {
        return this.unique;
    }
    
    public Gem validate(final ResourceLocation key) {
        Preconditions.checkNotNull((Object)this.dimensions);
        Preconditions.checkArgument(this.maxRarity.ordinal() >= this.minRarity.ordinal());
        RarityRegistry.INSTANCE.getValues().stream().filter(r -> r.isAtLeast(this.minRarity) && r.isAtMost(this.maxRarity)).forEach(r -> Preconditions.checkArgument(this.bonuses.stream().allMatch(b -> b.supports(r))));
        return this;
    }
    
    public Set<String> getStages() {
        return this.stages;
    }
    
    public Codec<? extends Gem> getCodec() {
        return Gem.CODEC;
    }
    
    public static void addTypeInfo(final Consumer<Component> list, final Object... types) {
        Arrays.sort(types, (c1, c2) -> ((LootCategory)c1).getName().compareTo(((LootCategory)c2).getName()));
        final Style style = Style.f_131099_.m_178520_(720650);
        if (types.length != LootCategory.BY_ID.size() - 1) {
            final StringBuilder sb = new StringBuilder();
            int rem;
            for (int i = 0; i < types.length; i += rem) {
                rem = Math.min(3, types.length - i);
                final Object[] args = new Object[rem];
                for (int r = 0; r < rem; ++r) {
                    sb.append("%s, ");
                    args[r] = Component.m_237115_(((LootCategory)types[i + r]).getDescIdPlural());
                }
                list.accept((Component)Component.m_237110_("text.apotheosis.dot_prefix", new Object[] { Component.m_237110_(sb.substring(), args) }).m_130948_(style));
                sb.setLength();
            }
        }
        else {
            list.accept((Component)Component.m_237110_("text.apotheosis.dot_prefix", new Object[] { Component.m_237115_("text.apotheosis.anything") }).m_130948_(style));
        }
    }
    
    public final ResourceLocation getId() {
        return GemRegistry.INSTANCE.getKey((CodecProvider)this);
    }
    
    static {
        CODEC = RecordCodecBuilder.create(inst -> inst.group((App)Codec.intRange(0, Integer.MAX_VALUE).fieldOf("weight").forGetter((Function)WeightedDynamicRegistry.ILuckyWeighted::getWeight), (App)PlaceboCodecs.nullableField(Codec.floatRange(0.0f, Float.MAX_VALUE), "quality", (Object)0.0f).forGetter((Function)WeightedDynamicRegistry.ILuckyWeighted::getQuality), (App)PlaceboCodecs.nullableField(PlaceboCodecs.setOf(ResourceLocation.f_135803_), "dimensions", (Object)Collections.emptySet()).forGetter((Function)WeightedDynamicRegistry.IDimensional::getDimensions), (App)PlaceboCodecs.nullableField(LootRarity.CODEC, "min_rarity").forGetter(g -> Optional.of(g.getMinRarity())), (App)PlaceboCodecs.nullableField(LootRarity.CODEC, "max_rarity").forGetter(g -> Optional.of(g.getMaxRarity())), (App)GemBonus.CODEC.listOf().fieldOf("bonuses").forGetter((Function)Gem::getBonuses), (App)PlaceboCodecs.nullableField((Codec)Codec.BOOL, "unique", (Object)false).forGetter((Function)Gem::isUnique), (App)PlaceboCodecs.nullableField(PlaceboCodecs.setOf((Codec)Codec.STRING), "stages").forGetter(gem -> Optional.ofNullable(gem.getStages()))).apply((Applicative)inst, Gem::new));
    }
}
